use std::any::Any;
use std::fmt::Display;
use std::sync::Arc;

use vortex_array::arrays::ConstantArray;
use vortex_array::{Array, ArrayRef};
use vortex_dtype::DType;
use vortex_error::VortexResult;
use vortex_scalar::Scalar;

use crate::{ExprRef, VortexExpr};

#[derive(Debug, PartialEq, Eq, Hash)]
pub struct Literal {
    value: Scalar,
}

impl Literal {
    pub fn new_expr(value: impl Into<Scalar>) -> ExprRef {
        Arc::new(Self {
            value: value.into(),
        })
    }

    pub fn value(&self) -> &Scalar {
        &self.value
    }

    pub fn maybe_from(expr: &ExprRef) -> Option<&Literal> {
        expr.as_any().downcast_ref::<Literal>()
    }
}

impl Display for Literal {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.value)
    }
}

#[cfg(feature = "proto")]
pub(crate) mod proto {
    use kind::Kind;
    use vortex_error::{VortexResult, vortex_bail, vortex_err};
    use vortex_proto::expr::kind;
    use vortex_scalar::Scalar;

    use crate::{ExprDeserialize, ExprRef, ExprSerializable, Id, Literal};

    pub(crate) struct LiteralSerde;

    impl Id for LiteralSerde {
        fn id(&self) -> &'static str {
            "literal"
        }
    }

    impl ExprDeserialize for LiteralSerde {
        fn deserialize(&self, kind: &Kind, _children: Vec<ExprRef>) -> VortexResult<ExprRef> {
            let Kind::Literal(value) = kind else {
                vortex_bail!("Expected literal kind");
            };
            let scalar: Scalar = value
                .value
                .as_ref()
                .ok_or_else(|| vortex_err!("empty literal scalar"))?
                .try_into()?;
            Ok(Literal::new_expr(scalar))
        }
    }

    impl ExprSerializable for Literal {
        fn id(&self) -> &'static str {
            LiteralSerde.id()
        }

        fn serialize_kind(&self) -> VortexResult<Kind> {
            Ok(Kind::Literal(kind::Literal {
                value: Some((&self.value).into()),
            }))
        }
    }
}

impl VortexExpr for Literal {
    fn as_any(&self) -> &dyn Any {
        self
    }

    fn unchecked_evaluate(&self, batch: &dyn Array) -> VortexResult<ArrayRef> {
        Ok(ConstantArray::new(self.value.clone(), batch.len()).into_array())
    }

    fn children(&self) -> Vec<&ExprRef> {
        vec![]
    }

    fn replacing_children(self: Arc<Self>, children: Vec<ExprRef>) -> ExprRef {
        assert_eq!(children.len(), 0);
        self
    }

    fn return_dtype(&self, _scope_dtype: &DType) -> VortexResult<DType> {
        Ok(self.value.dtype().clone())
    }
}

/// Create a new `Literal` expression from a type that coerces to `Scalar`.
///
///
/// ## Example usage
///
/// ```
/// use vortex_array::arrays::PrimitiveArray;
/// use vortex_dtype::Nullability;
/// use vortex_expr::{lit, Literal};
/// use vortex_scalar::Scalar;
///
/// let number = lit(34i32);
///
/// let literal = number.as_any()
///     .downcast_ref::<Literal>()
///     .unwrap();
/// assert_eq!(literal.value(), &Scalar::primitive(34i32, Nullability::NonNullable));
/// ```
pub fn lit(value: impl Into<Scalar>) -> ExprRef {
    Literal::new_expr(value.into())
}

#[cfg(test)]
mod tests {
    use std::sync::Arc;

    use vortex_dtype::{DType, Nullability, PType, StructDType};
    use vortex_scalar::Scalar;

    use crate::{lit, test_harness};

    #[test]
    fn dtype() {
        let dtype = test_harness::struct_dtype();

        assert_eq!(
            lit(10).return_dtype(&dtype).unwrap(),
            DType::Primitive(PType::I32, Nullability::NonNullable)
        );
        assert_eq!(
            lit(0_u8).return_dtype(&dtype).unwrap(),
            DType::Primitive(PType::U8, Nullability::NonNullable)
        );
        assert_eq!(
            lit(0.0_f32).return_dtype(&dtype).unwrap(),
            DType::Primitive(PType::F32, Nullability::NonNullable)
        );
        assert_eq!(
            lit(i64::MAX).return_dtype(&dtype).unwrap(),
            DType::Primitive(PType::I64, Nullability::NonNullable)
        );
        assert_eq!(
            lit(true).return_dtype(&dtype).unwrap(),
            DType::Bool(Nullability::NonNullable)
        );
        assert_eq!(
            lit(Scalar::null(DType::Bool(Nullability::Nullable)))
                .return_dtype(&dtype)
                .unwrap(),
            DType::Bool(Nullability::Nullable)
        );

        let sdtype = DType::Struct(
            Arc::new(StructDType::new(
                Arc::from([Arc::from("dog"), Arc::from("cat")]),
                vec![
                    DType::Primitive(PType::U32, Nullability::NonNullable),
                    DType::Utf8(Nullability::NonNullable),
                ],
            )),
            Nullability::NonNullable,
        );
        assert_eq!(
            lit(Scalar::struct_(
                sdtype.clone(),
                vec![Scalar::from(32_u32), Scalar::from("rufus".to_string())]
            ))
            .return_dtype(&dtype)
            .unwrap(),
            sdtype
        );
    }
}
